#!/bin/bash
# SPDX-License-Identifier: MIT

set -e

VGID="##VGID##"
PREBOOT="##PREBOOT##"
BOOT_OS_UUID="##BOOT_OS_UUID##"
ADMIN_USER="##ADMIN_USER##"

self="$0"
cd "${self%%step2.sh}"

system_dir="$(cd ../../../; pwd)"
os_name="${system_dir##*/}"

# clear
printf '\033[2J\033[H'

echo "Lernstick installer (second step)"
echo
echo "VGID: $VGID"
echo "System volume: $system_dir"
echo

BOLD="$(printf '\033[1m')"
RST="$(printf '\033[m')"

bputil -d -v "$VGID" >/tmp/bp.txt

if ! grep -q ': Paired' /tmp/bp.txt; then
    echo "Your system did not boot into the correct recoveryOS."
    echo
    echo "Each OS in your machine comes with its own copy of recoveryOS."
    echo "In order to complete the installation, we need to boot into"
    echo "the brand new recoveryOS that matches the OS which you are"
    echo "installing. The final installation step cannot be completed from"
    echo "a different recoveryOS."
    echo
    echo "Normally this should happen automatically after the initial"
    echo "installer sets up your new OS as the default boot option,"
    echo "but it seems something went wrong there. Let's try that again."
    echo
    echo "Press enter to continue."
    read

    while ! bless --setBoot --mount "$system_dir"; do
        echo
        echo "bless failed. Did you mistype your password?"
        echo "Press enter to try again."
        read
    done

    echo
    echo "Phew, hopefully that fixed it!"
    echo
    echo "Your system will now shut down. Once the screen goes blank,"
    echo "please wait 10 seconds, then press the power button and do not"
    echo "release it until you see the 'Entering startup options...'"
    echo "message, then select '$os_name' again."
    echo
    echo "Press enter to shut down your system."
    read
    shutdown -h now
    exit 1
fi

if ! grep -q 'one true recoveryOS' /tmp/bp.txt; then
    echo "Your system did not boot in One True RecoveryOS (1TR) mode."
    echo
    echo "To finish the installation, the system must be in this special"
    echo "mode. Perhaps you forgot to hold down the power button, or"
    echo "momentarily released it at some point?"
    echo
    echo "Note that tapping and then pressing the power button again will"
    echo "allow you to see the boot picker screen, but you will not be"
    echo "in the correct 1TR mode. You must hold down the power button"
    echo "in one continuous motion as you power on the machine."
    echo
    echo "Your system will now shut down. Once the screen goes blank,"
    echo "please wait 10 seconds, then press the power button and do not"
    echo "release it until you see the 'Entering startup options...'"
    echo "message, then select '$os_name' again."
    echo
    echo "Press enter to shut down your system."
    read
    shutdown -h now
    exit 1
fi
echo "We are now going through the final installation steps. You will"
echo "be asked to enter your password for the user \"$ADMIN_USER\" twice."
echo
echo "Press enter to continue."
echo

read

read -p "Enter password for user ${ADMIN_USER}: " -s PASSWORD
echo

TEMP_LOG=$(mktemp)
echo -n "Running bputil..."
while ! bputil -u "$ADMIN_USER" -p "$PASSWORD" -nc -v "$VGID" > "$TEMP_LOG" 2>&1 ; do
    echo
    cat "$TEMP_LOG"
    echo
    echo "bputil failed. Did you mistype your password?"
    read -p "Enter password for user ${ADMIN_USER}: " -s PASSWORD
    echo
    echo "Press enter to try again."
    read
done
echo "done"

echo

if [ -e "/System/Volumes/iSCPreboot/$VGID/boot" ]; then
    # This is an external volume, and kmutil has a problem with trying to pick
    # up the AdminUserRecoveryInfo.plist from the wrong place. Work around that.
    diskutil mount "$PREBOOT"
    preboot="$(diskutil info "$PREBOOT" | grep "Mount Point" | sed 's, *Mount Point: *,,')"
    cp -R "$preboot/$VGID/var" "/System/Volumes/iSCPreboot/$VGID/"
fi


echo "Running kmutil..."
while true; do
    FIFO=$(mktemp -u)
    mkfifo "$FIFO"
    exec 3<> "$FIFO"
    # We can already buffer the inputs, they are always the same
    # TODO add sanity checks before actually sending the inputs
    echo "y" >&3
    echo "$ADMIN_USER" >&3
    while IFS= read -d $'\0' -n 1 a ; do
        log+="${a}"
        str+="${a}"

        if [[ "${str}" =~ .*"Password:".* ]] ; then
            echo -n "Enter password for user $ADMIN_USER:"
            unset str
        fi

        if [ "$a" = $'\n' ] ; then
            unset str
        fi
    done < <(kmutil configure-boot -c boot.bin --raw --entry-point 2048 --lowest-virtual-address 0 -v "$system_dir" <$FIFO)

    if [ $? = 0 ]
    then
        rm "$FIFO"
        break
    else
        rm "$FIFO"
        echo
        printf log
        echo "kmutil failed. Did you mistype your password?"
        echo "Press enter to try again."
        read
    fi
done
echo "...done"

echo -n "Running bless..."
if [ "$BOOT_OS_UUID" != "" ]; then
  dev_path="$(diskutil info -plist $BOOT_OS_UUID | plutil -extract DeviceNode raw -)"
  while ! echo -n "$PASSWORD" | bless --setBoot --device "$dev_path" --user "$ADMIN_USER" --stdinpass > "$TEMP_LOG" 2>&1 ; do
          echo
          cat "$TEMP_LOG"
          echo
          echo "bless failed. Did you mistype your password?"
          read -p "Enter password for user ${ADMIN_USER}: " -s PASSWORD
          echo
          echo "Press enter to try again."
          read
      done
fi
echo "done"

echo
echo "Wrapping up..."
echo

mount -u -w "$system_dir"

if [ -e "$system_dir/.IAPhysicalMedia" ]; then
    mv "$system_dir/.IAPhysicalMedia" "$system_dir/IAPhysicalMedia-disabled.plist"
fi

if [ -e "$system_dir/System/Library/CoreServices/SystemVersion-disabled.plist" ]; then
    mv -f "$system_dir/System/Library/CoreServices/SystemVersion"{-disabled,}".plist"
fi

echo
echo "Installation complete! Press enter to reboot."
read

reboot
